import logging
import os
from dataclasses import dataclass
from pathlib import Path

from typing import Dict, List, Optional

from .training_plan import TrainingPlan

logger = logging.getLogger(__name__)

_REFERENCE_PLANS_DIR = Path(__file__).resolve().parent / "plans"


@dataclass(frozen=True)
class PlanInfo:
    plan_id: str
    name: str
    description: str


class PlanRepository:
    _plans: Dict[str, TrainingPlan] = {}

    _plan_infos: List[PlanInfo] = []

    _failed_plans: List[str] = []

    _include_reference_plans: bool = True

    _include_user_plans: bool = True

    @classmethod
    def get_plans(cls, refresh: bool = False, user_location: str = None) -> List[PlanInfo]:
        """
        Return all known plans.

        :param refresh: Forces an update of available plans from all repository sources.
        :param user_location: Can override user-defined plan location.  Generally not recommended for most use cases.
        """
        if refresh or len(cls._plans) == 0:
            cls._refresh_plans(user_location)

        return list(cls._plan_infos)

    def get_plan(self, plan_id: str, refresh_if_not_found: bool = True) -> Optional[TrainingPlan]:
        """
        Return plan by plan_id.  By default, this method update the list available plans from all repository sources if
        the requested plan is not found.

        :param plan_id: The plan id to get.
        :param refresh_if_not_found: Forces update of available plans from all repository sources if not found.
        """
        if plan_id not in self._plans and refresh_if_not_found:
            self._refresh_plans()

        return self._plans.get(plan_id)

    @classmethod
    def _refresh_plans(cls, user_location: str = None):
        cls._plans = {}
        cls._plan_infos = []
        cls._failed_plans = []

        if cls._include_reference_plans:
            cls._append_plan_paths(_get_reference_plan_files())

        if cls._include_user_plans:
            user_path = Path(user_location) if user_location is not None else None
            cls._append_plan_paths(_get_user_plan_paths(user_path))

    @classmethod
    def _append_plan_paths(cls, paths: List[Path]):
        for path in paths:
            try:
                plan = TrainingPlan.from_json_file(path)
                cls._plans[plan.plan_id] = plan
                cls._plan_infos.append(PlanInfo(plan.plan_id, plan.name, plan.description))
            except Exception as err:
                logger.error("Could not load plan %s: %s", path, err)
                cls._failed_plans.append(str(path))


def _get_reference_plan_files() -> List[Path]:
    return _generate_file_paths(_REFERENCE_PLANS_DIR)


def _get_user_plan_paths(location: Path = None) -> List[Path]:
    if location is None:
        location = Path.home().joinpath("Autotrainer").joinpath("training").joinpath("protocols")

    return _generate_file_paths(location)


def _generate_file_paths(location: Path) -> List[Path]:
    if Path(location).is_dir() and os.access(location, os.R_OK):
        try:
            return list(location.glob("*.json"))
        except FileNotFoundError:
            logger.warning(f"Training plans could not be read from {location}")

    return []



import logging
import sys
from pathlib import Path
from sys import float_info
from typing import List, Optional, Dict, Any

from autotrainer.core import get_verbose_logger
from typing_extensions import Self

from .training_protocols import (TrainingActionProtocol, TrainingPhaseProtocol, PredicateContext,
                                 class_from_dict, class_to_dict)
from .value_provider import BooleanValueProvider

logger = get_verbose_logger(__name__)

_FLOAT_COMPARISON = float_info.epsilon * 2


class TrainingAction(TrainingActionProtocol):
    @property
    def has_progress(self) -> bool:
        return False

    @property
    def is_complete(self) -> bool:
        return True

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext, is_init: bool) -> bool:
        return False

    def serialize_progress(self) -> Dict[str, Any]:
        return {}

    def deserialize_progress(self, data: Dict[str, Any]) -> None:
        pass

    def _to_dict(self) -> Optional[Dict[str, Any]]:
        return None

    def to_dict(self) -> Dict[str, Any]:
        return class_to_dict(self.__class__.__name__, self.__class__.__module__, self._to_dict())

    @classmethod
    def from_dict(cls, data: Optional[Dict[str, Any]]):
        return class_from_dict(data)


class HeadMagnetIntensityAction(TrainingAction):
    def __init__(self, start: float, increment: float, end: float, pellet_delta: int):
        # Parameters
        self.start = start
        self.increment = increment
        self.end = end
        self.pellet_delta = pellet_delta

        # Internal Progress
        self.pellet_start = 0
        self.current_intensity = 0

    @property
    def has_progress(self) -> bool:
        return True

    @property
    def is_complete(self) -> bool:
        return abs(self.current_intensity) >= self.end - _FLOAT_COMPARISON

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext, is_init: bool) -> bool:
        if is_init:
            self.pellet_start = context.progress.pellets_consumed
            self.current_intensity = self.start
            # Override any default phase setting
            if context.tunnel_device is not None:
                context.tunnel_device.update_head_magnet_intensity(self.start)
        else:
            if context.progress is None or context.tunnel_device is None:
                return False
            if context.progress.pellets_consumed - self.pellet_start > self.pellet_delta:
                current = context.tunnel_device.head_magnet_intensity
                if current is None:
                    current = self.current_intensity
                intensity = min(self.end, current + self.increment)
                self.current_intensity = intensity
                self.pellet_start = context.progress.pellets_consumed
                context.tunnel_device.update_head_magnet_intensity(intensity)

        return False

    def serialize_progress(self) -> Dict[str, Any]:
        return {
            "current_intensity": self.current_intensity,
            "pellet_start": self.pellet_start
        }

    def deserialize_progress(self, data: Dict[str, Any]) -> None:
        self.current_intensity = data["current_intensity"]
        self.pellet_start = data["pellet_start"]

    def _to_dict(self) -> Dict[str, Any]:
        return {
            "start": self.start,
            "increment": self.increment,
            "end": self.end,
            "pellet_delta": self.pellet_delta
        }

    @classmethod
    def from_properties(cls, data: Dict[str, Any]) -> Self:
        start = data["start"]
        increment = data["increment"]
        end = data["end"]
        pellet_delta = data["pellet_delta"]

        return HeadMagnetIntensityAction(start, increment, end, pellet_delta)


class ReachDistanceAction(TrainingAction):
    def __init__(self, increment: float, distance: float, pellet_delta: int):
        # Parameters
        self.increment = increment
        self.distance = distance
        self.pellet_delta = pellet_delta

        # Internal Progress
        self.current_increment_pellet_count = 0
        self.initial_y = 0
        self.current_delta_y = 0
        self.distance_reached_pellet_count = 0

    @property
    def has_progress(self) -> bool:
        return True

    @property
    def _has_reached_distance(self) -> bool:
        return abs(self.current_delta_y) >= self.distance - _FLOAT_COMPARISON

    @property
    def is_complete(self) -> bool:
        if not self._has_reached_distance:
            return False
        if self.distance_reached_pellet_count == 0:
            return False
        return self.current_increment_pellet_count - self.distance_reached_pellet_count >= self.pellet_delta

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext, is_init: bool) -> bool:
        if is_init:
            self.current_increment_pellet_count = context.progress.pellets_consumed
            if context.pellet_device is not None and context.pellet_device.last_dcs_set_position is not None:
                self.initial_y = context.pellet_device.last_dcs_set_position.y
            self.current_delta_y = 0
            self.distance_reached_pellet_count = 0
        else:
            if context.progress is None or context.pellet_device is None:
                return False
            if context.progress.pellets_consumed - self.current_increment_pellet_count > self.pellet_delta:
                self.current_increment_pellet_count = context.progress.pellets_consumed
                if not self._has_reached_distance:
                    last_position = context.pellet_device.last_dcs_set_position
                    if last_position is None:
                        return False
                    distance = min(self.initial_y + self.distance, last_position.y + self.increment)
                    self.current_delta_y = distance - self.initial_y
                    context.pellet_device.set_dcs_y(distance)
                    if self._has_reached_distance:
                        self.distance_reached_pellet_count = self.current_increment_pellet_count
                elif self.distance_reached_pellet_count == 0:
                    self.distance_reached_pellet_count = self.current_increment_pellet_count

        return False

    def serialize_progress(self) -> Dict[str, Any]:
        return {
            "current_increment_pellet_count": self.current_increment_pellet_count,
            "initial_y": self.initial_y,
            "current_delta_y": self.current_delta_y,
            "distance_reached_pellet_count": self.distance_reached_pellet_count
        }

    def deserialize_progress(self, data: Dict[str, Any]) -> None:
        self.current_increment_pellet_count = data.get("current_increment_pellet_count", 0)
        self.initial_y = data.get("initial_y", 0.0)
        self.current_delta_y = data.get("current_delta_y", 0.0)
        self.distance_reached_pellet_count = data.get("distance_reached_pellet_count", 0)

    def _to_dict(self) -> Dict[str, Any]:
        return {
            "distance": self.distance,
            "increment": self.increment,
            "pellet_delta": self.pellet_delta
        }

    @classmethod
    def from_properties(cls, data: Dict[str, Any]) -> Self:
        distance = data["distance"]
        increment = data["increment"]
        pellet_delta = data["pellet_delta"]

        return ReachDistanceAction(increment, distance, pellet_delta)


class ShuffledPropertyEnabledAction(TrainingAction):
    def __init__(self, property_path: str, value: bool, count: int, percentage: float):
        if not isinstance(property_path, list):
            self.property_path: List[str] = property_path.split(".")
        else:
            self.property_path = property_path
        self.value = value
        self.count = count
        self.percentage = percentage
        self._provider = BooleanValueProvider(value, count, percentage)

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext, is_init: bool) -> bool:
        target = context
        for part in self.property_path[:-1]:
            if target is not None and hasattr(target, part):
                target = getattr(target, part)
            else:
                return False

        if target is not None and hasattr(target, self.property_path[-1]):
            next_value = self._provider.get_next_value()
            setattr(target, self.property_path[-1], next_value)
        else:
            logger.warning(f"failed to find property {self.property_path[-1]} on target {target}")

        return False

    def _to_dict(self) -> Dict[str, Any]:
        return {
            "property_path": self.property_path,
            "value": self.value,
            "count": self.count,
            "percentage": self.percentage,
        }

    @classmethod
    def from_properties(cls, data: Dict[str, Any]) -> Self:
        property_path = data["property_path"]
        value = data["value"]
        count = data["count"]
        percentage = data["percentage"]

        return ShuffledPropertyEnabledAction(property_path, value, count, percentage)



import logging
import random
from typing import Generic, Tuple, TypeVar, Optional, List, Protocol, runtime_checkable

import numpy as np
from autotrainer.core import get_verbose_logger

T = TypeVar("T", covariant=True)
V = TypeVar("V")

logger = get_verbose_logger(__name__)


@runtime_checkable
class ValueProvider(Protocol[T]):

    def get_next_value(self) -> T: ...


class ShuffledValueProvider(Generic[V]):

    def __init__(self, value: V, count: Optional[int] = None, *, rng: Optional[random.Random] = None):
        self._rng = rng if rng is not None else random.Random()
        self._index = 0

        if count is not None:
            if count < 1:
                raise ValueError(f"'count' must be >= 1, got {count}.")

            self._values: Optional[List[V]] = self._generate_values(value, count)
            self._shuffle()
            self._constant = None
        else:
            self._values = None
            self._constant = value

    def _generate_values(self, value: V, count: int) -> List[V]:
        ...

    def _shuffle(self):
        self._rng.shuffle(self._values)
        self._index = 0

    def get_next_value(self) -> V:
        if self._constant is not None:
            return self._constant

        result = self._values[self._index]
        self._index += 1
        if self._index >= len(self._values):
            self._shuffle()
        return result


class BooleanValueProvider(ShuffledValueProvider[bool], ValueProvider[bool]):
    def __init__(self, value: bool, count: Optional[int] = None, percentage: Optional[float] = None, *,
                 rng: Optional[random.Random] = None):
        has_count = count is not None
        has_percentage = percentage is not None

        if has_count != has_percentage:
            raise ValueError(
                "Both 'count' and 'percentage' must be provided together, or neither."
            )

        if has_percentage and not (0.0 <= percentage <= 1.0):
            raise ValueError(f"'percentage' must be between 0.0 and 1.0, got {percentage}.")

        self._percentage = percentage
        super().__init__(value, count, rng=rng)

    def _generate_values(self, value: bool, count: int) -> List[bool]:
        value_count = round(count * self._percentage)
        opposite_count = count - value_count
        return [value] * value_count + [not value] * opposite_count


class FloatValueProvider(ShuffledValueProvider[float], ValueProvider[float]):
    def __init__(self, value: float, count: Optional[int] = None, value_range: Optional[Tuple[float, float]] = None,
                 *, rng: Optional[random.Random] = None):
        has_count = count is not None
        has_range = value_range is not None

        if has_count != has_range:
            raise ValueError(
                "Both 'count' and 'value_range' must be provided together, or neither."
            )

        self._interval: float = 0.0
        self._value_range = value_range
        super().__init__(value, count, rng=rng)

    @property
    def interval(self) -> float:
        return self._interval

    def _generate_values(self, value: float, count: int) -> List[float]:
        values, step = np.linspace(self._value_range[0], self._value_range[1], num=count, retstep=True)
        self._interval = float(step)
        return values.tolist()



from enum import IntEnum
from sys import float_info
from typing import Protocol, List, Union, Dict, Any, Optional

from typing_extensions import Self

from .training_protocols import PredicateContext, TrainingPhaseProtocol, class_from_dict, class_to_dict

_FLOAT_COMPARISON = float_info.epsilon * 2


class TrainingPredicate(Protocol):
    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext) -> bool:
        return False

    def _to_dict(self) -> Optional[Dict[str, Any]]:
        return None

    def to_dict(self) -> Dict[str, Any]:
        return class_to_dict(self.__class__.__name__, self.__class__.__module__, self._to_dict())

    @classmethod
    def from_dict(cls, data: Optional[Dict[str, Any]]):
        return class_from_dict(data)


class CompoundPredicate(TrainingPredicate):
    class CompoundType(IntEnum):
        AND = 0
        OR = 1

    def __init__(self, compound_type: CompoundType, predicates: List[TrainingPredicate]):
        self.compound_type = compound_type
        self.predicates = predicates

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext) -> bool:
        if self.compound_type == CompoundPredicate.CompoundType.AND:
            return all(predicate.evaluate(phase, context) for predicate in self.predicates)
        else:
            return any(predicate.evaluate(phase, context) for predicate in self.predicates)

    def _to_dict(self) -> Dict[str, Any]:
        return {
            "compound_type": self.compound_type.name,
            "predicates": [predicate.to_dict() for predicate in self.predicates]
        }

    @classmethod
    def from_properties(cls, data: Dict[str, Any]) -> Self:
        compound_type = CompoundPredicate.CompoundType[data["compound_type"]]
        predicates = [TrainingPredicate.from_dict(d) for d in data["predicates"]]

        return CompoundPredicate(compound_type, predicates)


class NumberComparisonPredicate(TrainingPredicate):
    class NumberComparisonType(IntEnum):
        LT = 0
        LTE = 1
        EQ = 2
        GTE = 3
        GT = 4

    def __init__(self, comparison_type: NumberComparisonType, path: str, value: Union[int, float]):
        self.comparison_type = comparison_type
        if not isinstance(path, list):
            self.path = path.split(".")
        else:
            self.path = path
        self.value = value

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext) -> bool:
        resolved = self._resolve_path(self.path, context)
        return self._compare(resolved)

    @staticmethod
    def _resolve_path(path: List[str], context: PredicateContext):
        eval_obj = context
        for part in path:
            if eval_obj and part:
                if hasattr(eval_obj, part):
                    eval_obj = eval_obj.__getattribute__(part)
                else:
                    return None

        return eval_obj

    def _compare(self, resolved_value) -> bool:
        if resolved_value is None:
            return False

        if self.comparison_type == NumberComparisonPredicate.NumberComparisonType.LT:
            return resolved_value < self.value
        elif self.comparison_type == NumberComparisonPredicate.NumberComparisonType.LTE:
            return resolved_value <= self.value
        elif self.comparison_type == NumberComparisonPredicate.NumberComparisonType.EQ:
            if isinstance(resolved_value, float) or isinstance(self.value, float):
                return abs(resolved_value - self.value) < _FLOAT_COMPARISON
            else:
                return resolved_value == self.value
        elif self.comparison_type == NumberComparisonPredicate.NumberComparisonType.GTE:
            return resolved_value >= self.value
        elif self.comparison_type == NumberComparisonPredicate.NumberComparisonType.GT:
            return resolved_value > self.value

        return False

    def _to_dict(self) -> Optional[Dict[str, Any]]:
        return {
            "comparison_type": self.comparison_type.name,
            "path": self.path,
            "value": self.value
        }

    @classmethod
    def from_properties(cls, data: Dict[str, Any]) -> Self:
        comparison_type = NumberComparisonPredicate.NumberComparisonType[data["comparison_type"]]
        path = data["path"]
        value = data["value"]

        return NumberComparisonPredicate(comparison_type, path, value)


class MathComparisonPredicate(NumberComparisonPredicate):
    class MathOperationType(IntEnum):
        ADD = 0
        SUBTRACT = 1
        MULTIPLY = 2
        DIVIDE = 3

    def __init__(self, comparison_type: NumberComparisonPredicate.NumberComparisonType,
                 left_path: str, math_operation: MathOperationType,
                 right_path: str, value: Union[int, float]):
        super().__init__(comparison_type, left_path, value)
        self.math_operation = math_operation
        if not isinstance(right_path, list):
            self.right_path = right_path.split(".")
        else:
            self.right_path = right_path

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext) -> bool:
        left_value = self._resolve_path(self.path, context)
        right_value = self._resolve_path(self.right_path, context)

        if left_value is None or right_value is None:
            return False

        if self.math_operation == MathComparisonPredicate.MathOperationType.ADD:
            result = left_value + right_value
        elif self.math_operation == MathComparisonPredicate.MathOperationType.SUBTRACT:
            result = left_value - right_value
        elif self.math_operation == MathComparisonPredicate.MathOperationType.MULTIPLY:
            result = left_value * right_value
        elif self.math_operation == MathComparisonPredicate.MathOperationType.DIVIDE:
            if right_value == 0:
                return False
            result = left_value / right_value
        else:
            return False

        return self._compare(result)

    def _to_dict(self) -> Optional[Dict[str, Any]]:
        return {
            "comparison_type": self.comparison_type.name,
            "left_path": self.path,
            "math_operation": self.math_operation.name,
            "right_path": self.right_path,
            "value": self.value
        }

    @classmethod
    def from_properties(cls, data: Dict[str, Any]) -> Self:
        comparison_type = NumberComparisonPredicate.NumberComparisonType[data["comparison_type"]]
        left_path = data["left_path"]
        math_operation = MathComparisonPredicate.MathOperationType[data["math_operation"]]
        right_path = data["right_path"]
        value = data["value"]

        return MathComparisonPredicate(comparison_type, left_path, math_operation, right_path, value)


class PythonPredicate(TrainingPredicate):
    def __init__(self, python_code: str):
        self.python_code = python_code

    def evaluate(self, phase: TrainingPhaseProtocol, context: PredicateContext) -> bool:
        pass

    def _to_dict(self) -> Optional[Dict[str, Any]]:
        return {
            "python_code": self.python_code
        }

    @classmethod
    def from_properties(cls, data: Dict[str, Any]):
        python_code = data["python_code"]
        return PythonPredicate(python_code)
